
/*

SSProxy 1.2 by Grelminar (grelminar@hotmail.com)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

(see http://www.gnu.org/)

*/


#define CONTROLPORT 15000
#define STARTPORT 5000
#define MAPNUM 128
#define MAXCLIENTS 10


#ifdef WINDOWS

#define FD_SETSIZE (MAPNUM+MAXCLIENTS+3)

#include <winsock.h>

#else

#include <netinet/in.h>
#include <sys/time.h>

#endif

#include <stdio.h>



#define IPEQUAL(a,b) (a.sin_addr.s_addr==b.sin_addr.s_addr)
#define PORTEQUAL(a,b) (a.sin_port==b.sin_port)
#define SINEQUAL(a,b) (IPEQUAL(a,b) && PORTEQUAL(a,b))


#ifdef WINDOWS
typedef SOCKET socket_t;
typedef unsigned int ip_t;
typedef unsigned short port_t;
#else
typedef int socket_t;
typedef uint32_t ip_t;
typedef uint16_t port_t;
#endif


struct Addr
{ ip_t ip; port_t port; };

struct Entry
{
	struct sockaddr_in dest;
	struct sockaddr_in client;
};

struct Client
{
	struct sockaddr_in sin;
	struct Entry map[MAPNUM];
	socket_t sender;
};

struct Client clients[MAXCLIENTS];
int nextc;
socket_t lnr[MAPNUM];




void Error(char *s)
{
#ifdef WINDOWS
	fprintf(stderr,"Error: %s (%i)\n",s,WSAGetLastError());
#else
	fprintf(stderr,"Error: %s\n",s);
#endif
	exit(1);
}


void MakeSin(struct sockaddr_in *sin, ip_t ip, port_t port)
{
	memset(sin->sin_zero,0,sizeof(sin->sin_zero));
	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = ip;
	sin->sin_port = port;
}
	

socket_t InitSocket(port_t port)
{
	socket_t s;
	struct sockaddr_in sin;

	MakeSin(&sin, INADDR_ANY, htons(port));

	if ((s = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
		Error("socket failed");
	if (bind(s, &sin, sizeof(sin)) == -1)
		Error("bind failed");
	return s;
}

void CloseSocket(socket_t s)
{
#ifdef WINDOWS
	closesocket(s);
#else
	close(s);
#endif
}


int ProcessControlData(char *b, int l, struct sockaddr_in *sin, int sinsize)
{
	static int nextc = 0;
	int i, mapc, c;
	struct Addr *a;
	
	if ((l % 6) || (l > (6*MAPNUM)))
	{
		fprintf(stderr,"Bad data on control port\n");
		return 0;
	}

	c = nextc++;
	for (i = 0; i < MAXCLIENTS; i++)
		if (IPEQUAL((*sin), clients[i].sin))
		{ c = i; nextc--; break; }
	nextc = nextc % MAXCLIENTS;

	printf("New connection incoming, assigning slot %i\n", c);

	if (!clients[c].sender) clients[c].sender = InitSocket(0);

	mapc = l / 6;

	for (i = 0; i < mapc; i++)
	{
		a = (struct Addr*)(b+i*6);
		MakeSin(&clients[c].map[i].dest, a->ip, a->port);
		printf("Map: %i(%i) -> %s:%i\n", STARTPORT+i, c,
					inet_ntoa(clients[c].map[i].dest.sin_addr), ntohs(a->port));
	}

	memcpy(&clients[c].sin, sin, sinsize);

	return 1;
}



int main()
{
	char buf[512];
	int i, j, ret, sinsize, maxfd, magic = 0x12345678;
	socket_t ctrl;
	fd_set fds;
	struct timeval tv;
	struct sockaddr_in sin;

#ifdef WINDOWS
	WSADATA wsad;

	WSAStartup(257,&wsad);
#endif

	printf("SSProxy 1.2 by Grelminar (grelminar@hotmail.com)\n");

	for (i = 0; i < MAXCLIENTS; i++)
		clients[i].sender = 0;

	ctrl = InitSocket(CONTROLPORT);
	for (i = 0; i < MAPNUM; i++)
		lnr[i] = InitSocket(STARTPORT+i);

	for ( ; ; )
	{
		FD_ZERO(&fds);
		
		FD_SET(ctrl, &fds); maxfd = ctrl;
		for (i = 0; i < MAXCLIENTS; i++)
		{
			if (clients[i].sender) FD_SET(clients[i].sender, &fds);
			if (clients[i].sender > maxfd) maxfd = clients[i].sender;
		}
		for (i = 0; i < MAPNUM; i++)
		{
			FD_SET(lnr[i], &fds);
			if (lnr[i] > maxfd) maxfd = lnr[i];
		}

		tv.tv_sec = 10; tv.tv_usec = 0;
		ret = select(maxfd+1, &fds, NULL, NULL, &tv);
		if (ret == -1) Error("select");

		for (i = 0; i < MAPNUM; i++)
			if (FD_ISSET(lnr[i], &fds))
			{
				//printf("data on listener %i ",i); fflush(stdout);
				sinsize = sizeof(sin);
				ret = recvfrom(lnr[i], buf, 512, 0, &sin, &sinsize);
				for (j = 0; j < MAXCLIENTS; j++)
					if (IPEQUAL(sin, clients[j].sin))
					{
						//printf("from client %i\n",j);
						sendto(clients[j].sender, buf, ret, 0, 
							&clients[j].map[i].dest, sizeof(struct sockaddr_in));
						memcpy(&clients[j].map[i].client, &sin, sinsize);
						break;
					}
			}
		for (j = 0; j < MAXCLIENTS; j++)
			if (FD_ISSET(clients[j].sender, &fds))
			{
				//printf("data on sender %i ",j); fflush(stdout);
				sinsize = sizeof(sin);
				ret = recvfrom(clients[j].sender, buf, 512, 0, &sin, &sinsize);
				for (i = 0; i < MAPNUM; i++)
					if (SINEQUAL(sin, clients[j].map[i].dest))
					{
						//printf("for mapping %i\n",i);
						sendto(lnr[i], buf, ret, 0,
							&clients[j].map[i].client, sizeof(struct sockaddr_in));
						break;
					}
			}
		if (FD_ISSET(ctrl, &fds))
		{
			sinsize = sizeof(sin);
			ret = recvfrom(ctrl, buf, 512, 0, &sin, &sinsize);
			if (ProcessControlData(buf, ret, &sin, sinsize))
				sendto(ctrl, &magic, 4, 0, &sin, sinsize);
		}
	}
	
	// this will never run, but it's nice to know it's here
 
	for (i = 0; i < MAPNUM; i++)
		CloseSocket(lnr[i]);
	for (i = 0; i < MAXCLIENTS; i++)
		if (clients[i].sender) CloseSocket(clients[i].sender);
	CloseSocket(ctrl);

#ifdef WINDOWS
	WSACleanup();
#endif

	return 0;
}



